/*-
 * Copyright (c) 2007, Bjrn Knig
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/systm.h>
#include <sys/exec.h>
#include <sys/imgact.h>
#include <sys/kernel.h>

#if BYTE_ORDER == LITTLE_ENDIAN
#define MZMAGIC		0x5a4d	/* MZ */
#define PEMAGIC		0x4550	/* PE */
#else
#define MZMAGIC		0x4d5a
#define PEMAGIC		0x5045
#endif

int mono_enabled = 1;
char mono_path[MAXPATHLEN] = "/usr/local/bin/mono";
char mono_options[255] = "";

SYSCTL_NODE(_kern, OID_AUTO, mono, CTLFLAG_RW,
	0, "Mono magic");
SYSCTL_INT(_kern_mono, OID_AUTO, enabled, CTLFLAG_RW,
	&mono_enabled, 0, "To be enabled or not to be enabled, that is the question.");
SYSCTL_STRING(_kern_mono, OID_AUTO, path, CTLFLAG_RW,
	mono_path, sizeof mono_path, "Path to the .NET runtime interpreter");
SYSCTL_STRING(_kern_mono, OID_AUTO, options, CTLFLAG_RW,
	mono_options, sizeof mono_options, "Options for the .NET runtime interpreter");

static int exec_mono_imgact(struct image_params *imgp);

static int
exec_mono_imgact(imgp)
	struct image_params *imgp;
{
	size_t length;
	int32_t offset;
	uint32_t PEHeaderOffset, CLRHeaderSize;

	if (!mono_enabled)
		return(-1);

	if (((const short *) imgp->image_header)[0] != MZMAGIC)
		return(-1);

	PEHeaderOffset = ((const uint32_t *) imgp->image_header)[0x3c / sizeof(uint32_t)];

	if ((((const uint16_t *) imgp->image_header)[PEHeaderOffset / sizeof(uint16_t)]) != PEMAGIC)
		return(ENOEXEC);

	CLRHeaderSize = ((const uint32_t *) imgp->image_header)[(PEHeaderOffset + 4 + 20 + 208 + 4) / sizeof(uint32_t)];

	if (CLRHeaderSize == 0)
		return (ENOEXEC);

	if (imgp->interpreted)
		return (ENOEXEC);

	imgp->interpreted = 1;

	offset = strlen(mono_path) + 1;
	if (strlen(mono_options) > 0)
		offset += strlen(mono_options) + 1;

	if (offset > imgp->args->stringspace)
		return (E2BIG);

	bcopy(imgp->args->begin_argv, imgp->args->begin_argv + offset, imgp->args->endp - imgp->args->begin_argv);
	
	imgp->args->begin_envv += offset;
	imgp->args->endp += offset;
	imgp->args->stringspace -= offset;

	imgp->args->argc++;
	bcopy(mono_path, imgp->args->begin_argv, strlen(mono_path));
	imgp->args->begin_argv[strlen(mono_path)] = '\0';

	if (strlen(mono_options) > 0)
	{
		imgp->args->argc++;
		bcopy(mono_options, imgp->args->begin_argv + strlen(mono_path) + 1, strlen(mono_options));
		imgp->args->begin_argv[strlen(mono_path) + 1 + strlen(mono_options)] = '\0';
		
		// replace all occurences of ' ' with '\0'
		for (char *p = imgp->args->begin_argv + strlen(mono_path) + 1; *p != '\0'; p++)
			if (*p == ' ')
			{
				imgp->args->argc++;
				*p = '\0';
			}
	}

	copystr(mono_path, imgp->interpreter_name, strlen(mono_path), &length);

	return (0);
}

static struct execsw mono_execsw = { exec_mono_imgact, "MZ" };
EXEC_SET(mono, mono_execsw);

